#include <stdio.h>
#include <stddef.h>
#include <assert.h>
#include <stdlib.h>
#include "ufbx.h"

void create_vertex_buffer(const void *vertices, size_t count)
{
    printf(".. %zu vertices\n", count);
}

void create_index_buffer(const uint32_t *indices, size_t count)
{
    printf(".. %zu indices\n", count);
}

// GPU vertex representation.
// In practice you would need to use more compact custom vector types as
// by default `ufbx_real` is 64 bits wide.
typedef struct Vertex {
    ufbx_vec3 position;
    ufbx_vec3 normal;
    ufbx_vec2 uv;
} Vertex;

void convert_mesh_part(ufbx_mesh *mesh, ufbx_mesh_part *part)
{
    size_t num_triangles = part->num_triangles;
    Vertex *vertices = calloc(num_triangles * 3, sizeof(Vertex));
    size_t num_vertices = 0;

    // Reserve space for the maximum triangle indices.
    size_t num_tri_indices = mesh->max_face_triangles * 3;
    uint32_t *tri_indices = calloc(num_tri_indices, sizeof(uint32_t));

    // Iterate over each face using the specific material.
    for (size_t face_ix = 0; face_ix < part->num_faces; face_ix++) {
        ufbx_face face = mesh->faces.data[part->face_indices.data[face_ix]];

        // Triangulate the face into `tri_indices[]`.
        uint32_t num_tris = ufbx_triangulate_face(tri_indices, num_tri_indices, mesh, face);

        // Iterate over each triangle corner contiguously.
        for (size_t i = 0; i < num_tris * 3; i++) {
            uint32_t index = tri_indices[i];

            Vertex *v = &vertices[num_vertices++];
            v->position = ufbx_get_vertex_vec3(&mesh->vertex_position, index);
            v->normal = ufbx_get_vertex_vec3(&mesh->vertex_normal, index);
            v->uv = ufbx_get_vertex_vec2(&mesh->vertex_uv, index);
        }
    }

    // Should have written all the vertices.
    free(tri_indices);
    assert(num_vertices == num_triangles * 3);

    // Generate the index buffer.
    ufbx_vertex_stream streams[1] = {
        { vertices, num_vertices, sizeof(Vertex) },
    };
    size_t num_indices = num_triangles * 3;
    uint32_t *indices = calloc(num_indices, sizeof(uint32_t));

    // This call will deduplicate vertices, modifying the arrays passed in `streams[]`,
    // indices are written in `indices[]` and the number of unique vertices is returned.
    num_vertices = ufbx_generate_indices(streams, 1, indices, num_indices, NULL, NULL);

    create_vertex_buffer(vertices, num_vertices);
    create_index_buffer(indices, num_indices);

    free(indices);
    free(vertices);
}

int main(int argc, char **argv)
{
    ufbx_scene *scene = ufbx_load_file("material_parts.fbx", NULL, NULL);
    assert(scene);

    for (size_t mesh_ix = 0; mesh_ix < scene->meshes.count; mesh_ix++) {
        ufbx_mesh *mesh = scene->meshes.data[mesh_ix];
        printf("mesh '%s'\n", mesh->name.data);

        for (size_t part_ix = 0; part_ix < mesh->material_parts.count; part_ix++) {
            ufbx_mesh_part *part = &mesh->material_parts.data[part_ix];
            ufbx_material *material = NULL;
            if (part_ix < mesh->materials.count) {
                material = mesh->materials.data[part_ix];
            }
            printf(". [%zu] material: %s\n", part_ix,
                material ? material->name.data : "(none)");
            convert_mesh_part(mesh, part);
        }
    }

    return 0;
}
